//! The instruction architecture of the `wasmi` interpreter.

mod utils;

#[cfg(test)]
mod tests;

pub use self::utils::{DropKeep, FuncIdx, GlobalIdx, LocalIdx, Offset, SignatureIdx, Target};
use wasmi_core::UntypedValue;

/// The internal `wasmi` bytecode that is stored for Wasm functions.
///
/// # Note
///
/// This representation slightly differs from WebAssembly instructions.
///
/// For example the `BrTable` instruciton is unrolled into separate instructions
/// each representing either the `BrTable` head or one of its branching targets.
#[derive(Copy, Debug, Clone, PartialEq, Eq)]
pub enum Instruction {
    LocalGet {
        local_depth: LocalIdx,
    },
    /// Variant of `LocalGet` required to differentiate
    /// between empty and non-empty value stacks.
    LocalGetEmpty {
        local_depth: LocalIdx,
    },
    LocalSet {
        local_depth: LocalIdx,
    },
    LocalTee {
        local_depth: LocalIdx,
    },
    Br(Target),
    /// Variant of `Br` required to differentiate
    /// between empty and non-empty value stacks.
    BrEmpty(Target),
    BrIfEqz(Target),
    BrIfNez(Target),
    ReturnIfNez(DropKeep),
    BrTable {
        len_targets: usize,
    },
    Unreachable,
    Return(DropKeep),
    /// Variant of `Return` required to differentiate
    /// between empty and non-empty value stacks.
    ReturnEmpty(DropKeep),
    Call(FuncIdx),
    /// Variant of `Call` required to differentiate
    /// between empty and non-empty value stacks.
    CallEmpty(FuncIdx),
    CallIndirect(SignatureIdx),
    Drop,
    Select,
    GlobalGet(GlobalIdx),
    /// Variant of `GlobalGet` required to differentiate
    /// between empty and non-empty value stacks.
    GlobalGetEmpty(GlobalIdx),
    GlobalSet(GlobalIdx),
    I32Load(Offset),
    I64Load(Offset),
    F32Load(Offset),
    F64Load(Offset),
    I32Load8S(Offset),
    I32Load8U(Offset),
    I32Load16S(Offset),
    I32Load16U(Offset),
    I64Load8S(Offset),
    I64Load8U(Offset),
    I64Load16S(Offset),
    I64Load16U(Offset),
    I64Load32S(Offset),
    I64Load32U(Offset),
    I32Store(Offset),
    I64Store(Offset),
    F32Store(Offset),
    F64Store(Offset),
    I32Store8(Offset),
    I32Store16(Offset),
    I64Store8(Offset),
    I64Store16(Offset),
    I64Store32(Offset),
    MemorySize,
    /// Variant of `MemorySize` required to differentiate
    /// between empty and non-empty value stacks.
    MemorySizeEmpty,
    MemoryGrow,
    Const(UntypedValue),
    /// Variant of `Const` required to differentiate
    /// between empty and non-empty value stacks.
    ConstEmpty(UntypedValue),
    I32Eqz,
    I32Eq,
    I32Ne,
    I32LtS,
    I32LtU,
    I32GtS,
    I32GtU,
    I32LeS,
    I32LeU,
    I32GeS,
    I32GeU,
    I64Eqz,
    I64Eq,
    I64Ne,
    I64LtS,
    I64LtU,
    I64GtS,
    I64GtU,
    I64LeS,
    I64LeU,
    I64GeS,
    I64GeU,
    F32Eq,
    F32Ne,
    F32Lt,
    F32Gt,
    F32Le,
    F32Ge,
    F64Eq,
    F64Ne,
    F64Lt,
    F64Gt,
    F64Le,
    F64Ge,
    I32Clz,
    I32Ctz,
    I32Popcnt,
    I32Add,
    I32Sub,
    I32Mul,
    I32DivS,
    I32DivU,
    I32RemS,
    I32RemU,
    I32And,
    I32Or,
    I32Xor,
    I32Shl,
    I32ShrS,
    I32ShrU,
    I32Rotl,
    I32Rotr,
    I64Clz,
    I64Ctz,
    I64Popcnt,
    I64Add,
    I64Sub,
    I64Mul,
    I64DivS,
    I64DivU,
    I64RemS,
    I64RemU,
    I64And,
    I64Or,
    I64Xor,
    I64Shl,
    I64ShrS,
    I64ShrU,
    I64Rotl,
    I64Rotr,
    F32Abs,
    F32Neg,
    F32Ceil,
    F32Floor,
    F32Trunc,
    F32Nearest,
    F32Sqrt,
    F32Add,
    F32Sub,
    F32Mul,
    F32Div,
    F32Min,
    F32Max,
    F32Copysign,
    F64Abs,
    F64Neg,
    F64Ceil,
    F64Floor,
    F64Trunc,
    F64Nearest,
    F64Sqrt,
    F64Add,
    F64Sub,
    F64Mul,
    F64Div,
    F64Min,
    F64Max,
    F64Copysign,
    I32WrapI64,
    I32TruncSF32,
    I32TruncUF32,
    I32TruncSF64,
    I32TruncUF64,
    I64ExtendSI32,
    I64ExtendUI32,
    I64TruncSF32,
    I64TruncUF32,
    I64TruncSF64,
    I64TruncUF64,
    F32ConvertSI32,
    F32ConvertUI32,
    F32ConvertSI64,
    F32ConvertUI64,
    F32DemoteF64,
    F64ConvertSI32,
    F64ConvertUI32,
    F64ConvertSI64,
    F64ConvertUI64,
    F64PromoteF32,
    I32ReinterpretF32,
    I64ReinterpretF64,
    F32ReinterpretI32,
    F64ReinterpretI64,
    I32Extend8S,
    I32Extend16S,
    I64Extend8S,
    I64Extend16S,
    I64Extend32S,
    I32TruncSatF32S,
    I32TruncSatF32U,
    I32TruncSatF64S,
    I32TruncSatF64U,
    I64TruncSatF32S,
    I64TruncSatF32U,
    I64TruncSatF64S,
    I64TruncSatF64U,
}

impl Instruction {
    /// Creates a new `Const` instruction from the given value.
    pub fn constant<T>(value: T) -> Self
    where
        T: Into<UntypedValue>,
    {
        Self::Const(value.into())
    }

    /// Creates a new `Const` instruction from the given value.
    ///
    /// # Note
    ///
    /// This is a special variant indicating that this operation acts on
    /// an empty value stack.
    pub fn constant_empty<T>(value: T) -> Self
    where
        T: Into<UntypedValue>,
    {
        Self::ConstEmpty(value.into())
    }

    /// Creates a new `local.get` instruction from the given local depth.
    pub fn local_get(local_depth: u32) -> Self {
        Self::LocalGet {
            local_depth: LocalIdx::from(local_depth),
        }
    }

    /// Creates a new `local.get` instruction from the given local depth.
    ///
    /// # Note
    ///
    /// This is a special variant indicating that this operation acts on
    /// an empty value stack.
    pub fn local_get_empty(local_depth: u32) -> Self {
        Self::LocalGetEmpty {
            local_depth: LocalIdx::from(local_depth),
        }
    }

    /// Creates a new `local.set` instruction from the given local depth.
    pub fn local_set(local_depth: u32) -> Self {
        Self::LocalSet {
            local_depth: LocalIdx::from(local_depth),
        }
    }

    /// Creates a new `local.tee` instruction from the given local depth.
    pub fn local_tee(local_depth: u32) -> Self {
        Self::LocalTee {
            local_depth: LocalIdx::from(local_depth),
        }
    }
}
